# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2021-08-07 19:07
from hanlp_common.constant import HANLP_URL

CPB3_SRL_ELECTRA_SMALL = HANLP_URL + 'srl/cpb3_electra_small_crf_20210807_190553.zip'
'Electra small model (:cite:`clark2020electra`) trained on CPB3. F1 = `76.53`.'

ALL = {}
