# -*- coding:utf-8 -*-
# Author: hankcs
# Date: 2019-12-30 18:57
FASTTEXT_DEBUG_EMBEDDING_EN = 'https://elit-models.s3-us-west-2.amazonaws.com/fasttext.debug.bin.zip'
FASTTEXT_CC_300_EN = 'https://dl.fbaipublicfiles.com/fasttext/vectors-crawl/cc.en.300.bin.gz'
'FastText (:cite:`bojanowski2017enriching`) embeddings trained on Common Crawl.'
FASTTEXT_WIKI_NYT_AMAZON_FRIENDS_200_EN \
    = 'https://elit-models.s3-us-west-2.amazonaws.com/fasttext-200-wikipedia-nytimes-amazon-friends-20191107.bin'
'FastText (:cite:`bojanowski2017enriching`) embeddings trained on wikipedia, nytimes and friends.'

FASTTEXT_WIKI_300_ZH = 'https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.zh.zip#wiki.zh.bin'
'FastText (:cite:`bojanowski2017enriching`) embeddings trained on Chinese Wikipedia.'
FASTTEXT_WIKI_300_ZH_CLASSICAL = 'https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.zh_classical.zip#wiki.zh_classical.bin'
'FastText (:cite:`bojanowski2017enriching`) embeddings trained on traditional Chinese wikipedia.'

ALL = {}
